#include "common.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <termios.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>
#include <locale.h>
#include <sys/types.h>
#include <pwd.h>
#include <dirent.h>
#include <stdlib.h>
#include <fcntl.h>

void P(char* msg, ...);

BOOL gMainLoop = FALSE;
void (*gView)() = NULL;          // 登録描写関数
void (*gView2)() = NULL;          // 登録描写関数

///////////////////////////////////////////////////////////////////////////////
// キーボード入力処理
///////////////////////////////////////////////////////////////////////////////
static void input(int meta, int key)
{
    if(key == 26) {
        mendwin();
        kill(getpid(), SIGSTOP);
        minitscr();
    }
    else if(gCmdLineActive) {
        /// 補完候補選択 ///
        if(vector_size(gCCandidate) > 0)
            cmdline_completion_input(meta, key);

        /// 普通のコマンドライン ///
        else
            cmdline_input(meta, key);
    }
    else if(gActiveMenu) {
        /// メニュー ///
        menu_input(gActiveMenu, meta, key);
    }
    /// インクリメンタルサーチ ///
    else if(gISearch) {
        isearch_input(meta, key);
    }
    else {
        /// ファイラー ///
        filer_input(meta, key);
    }
}

///////////////////////////////////////////////////////////////////////////////
// 描写
///////////////////////////////////////////////////////////////////////////////
void view()
{
    if(gCmdLineActive) {
        /// 補完候補選択画面 ///
        if(vector_size(gCCandidate)) {
            cmdline_completion_view();
            cmdline_view();
        }
        /// ヒストリー選択画面 ///
        else if(vector_size(gHCandidate)) {
            cmdline_history_view();
            cmdline_view();
        }

        /// ファイラ画面 ///
        else {
            filer_view(0);
            filer_view(1);

            shell_view();

            cmdline_view();
        }
    }
    else if(gISearch) {
        filer_view(0);
        filer_view(1);

        shell_view();

        isearch_view();
    }
    else if(gActiveMenu) {
        filer_view(0);
        filer_view(1);

        shell_view();

        cmdline_view();
        menu_view(gActiveMenu);
    }
    else {
        filer_view(0);
        filer_view(1);

        shell_view();
        cmdline_view();
    }

    if(gView) gView();
    if(gView2) gView2();
}

///////////////////////////////////////////////////////////////////////////////
// ランタイムスクリプト実行
///////////////////////////////////////////////////////////////////////////////
static void read_rc_file()
{
    setenv("DATADIR", DATADIR, 1);

    /* ~/.mfiler3を読み込む。もし ~/.mfiler3がないなら$prefix/etc/.mfiler3を読み込む */
    char* home = getenv("HOME");
    if(home == NULL) {
        fprintf(stderr, "$HOME is NULL");
        exit(1);
    }

    char rc_fname[256];
    sprintf(rc_fname, "%s/.mfiler3", home);
    
    if(access(rc_fname, R_OK) == 0) {
        shell_load(rc_fname, TRUE);
    }
    else {
        sprintf(rc_fname,"%s/.mfiler3", SYSCONFDIR);
        if(access(rc_fname, R_OK) == 0) {
            shell_load(rc_fname, TRUE);
        }
        else {
            fprintf(stderr, "can't find $HOME/.mfiler3 or %s/.mfiler3 file\n", SYSCONFDIR);
            exit(1);
        }
    }


}

///////////////////////////////////////////////////////////////////////////////
// exit時に実行される関数
///////////////////////////////////////////////////////////////////////////////
static void atexit_fun()
{
    if(mis_curses()) {
        mendwin();
    }
}

///////////////////////////////////////////////////////////////////////////////
// シグナル処理
///////////////////////////////////////////////////////////////////////////////
static void sig_winch(int signal)
{
    if(gMainLoop) {
        mclear_immediately();       // 画面の再描写
        view();
        mrefresh();
    }
}


void sig_SIGQUIT() { P("SIGQUIT"); }
void sig_SIGCHLD() { P("SIGCHLD"); }
void sig_SIGABRT() { P("SIGABRT"); }
void sig_SIGALRM() { P("SIGALRM"); }
void sig_SIGBUS() { P("SIGBUS"); }
void sig_SIGCONT() { P("SIGCONT"); }
void sig_SIGHUP() { P("SIGHUP"); }
void sig_SIGFPE() { P("SIGFPE"); }
void sig_SIGILL() { P("SIGILL"); }
void sig_SIGIO() { P("SIGIO"); }
void sig_SIGKILL() { P("SIGKILL"); }
void sig_SIGPIPE() { P("SIGPIPE"); }
void sig_SIGPROF() { P("SIGPROF"); }
void sig_SIGSEGV() { P("SIGSEGV"); }
void sig_SIGTERM() { P("SIGTERM"); }
void sig_SIGTRAP() { P("SIGTRAP"); }
void sig_SIGURG() { P("SIGURG"); }
void sig_SIGUSR1() { P("SIGUSR1"); }
void sig_SIGUSR2() { P("SIGUSR2"); }
void sig_SIGXCPU() { P("SIGXCPU"); }
void sig_SIGXFSZ() { P("SIGXFSZ"); }
void sig_SIGVTALRM() { P("SIGVTALRM"); }
void sig_SIGPOLL() { P("SIGPOLL"); }
void sig_SIGSYS() { P("SIGSYS"); }
void sig_SIGUNUSED() { P("SIGUNUSED"); }

void sig_SIGTTIN() { P("SIGTTIN"); }
void sig_SIGTTOU() { P("SIGTTOU"); }
void sig_SIGWINCH() { P("SIGWINCH"); }
void sig_SIGSTOP() { P("SIGSTOP"); }
void sig_SIGTSTP() { P("SIGTSTP"); }

int gBreak = FALSE;
void sig_SIGINT() { gBreak = TRUE; }

void set_signal_sigint()
{
    signal(SIGINT, sig_SIGINT);
}

void set_signal_sigint_clear()
{
    signal(SIGINT, SIG_IGN);
}

void set_signal_clear_all()
{
    signal(SIGABRT, SIG_DFL);
    signal(SIGALRM, SIG_DFL);
    signal(SIGBUS, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);
    signal(SIGCONT, SIG_DFL);
    signal(SIGHUP, SIG_DFL);
    signal(SIGFPE, SIG_DFL);
    signal(SIGILL, SIG_DFL);
    signal(SIGINT, SIG_DFL);
    signal(SIGIO, SIG_DFL);
    signal(SIGKILL, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGPROF, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGTRAP, SIG_DFL);
    signal(SIGTTIN, SIG_DFL);
    signal(SIGTTOU, SIG_DFL);
    signal(SIGWINCH, SIG_DFL);
    signal(SIGURG, SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    signal(SIGUSR2, SIG_DFL);
    signal(SIGXCPU, SIG_DFL);
    signal(SIGXFSZ, SIG_DFL);
    signal(SIGVTALRM, SIG_DFL);
    signal(SIGSYS, SIG_DFL);
#if !defined __CYGWIN__
    signal(SIGUNUSED, SIG_DFL);
#endif
}


void set_signal_clear()
{
    signal(SIGWINCH, SIG_DFL);
    signal(SIGTTOU, SIG_DFL);
    signal(SIGTTIN, SIG_DFL);
    signal(SIGINT, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGSTOP, SIG_DFL);
    signal(SIGTSTP, SIG_DFL);
}

void set_signal_subshell()
{
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGSTOP, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGINT, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
}

void set_signal_mfiler3()
{
    signal(SIGWINCH, sig_winch);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGSTOP, SIG_DFL);
    signal(SIGTSTP, SIG_IGN);
}

///////////////////////////////////////////////////////////////////////////////
// メイン関数
///////////////////////////////////////////////////////////////////////////////
char gTempDir[PATH_MAX];
char gHomeDir[PATH_MAX];

int main(int argc, char* argv[])
{
    /// ホームディレクトリを保存しておく ///
    char* home_dir = getenv("HOME");
    if(home_dir == NULL) {
        fprintf(stderr, "set $HOME");
        exit(1);
    }
    strcpy(gHomeDir, home_dir);

    /// 一時ファイルの置き場を保存しておく ///
    sprintf(gTempDir, "%s/.mfiler3_tmp", gHomeDir);
    if(access(gTempDir, F_OK) != 0) {
        if(mkdir(gTempDir, 0700) < 0) {
            fprintf(stderr, "can't mkdir %s", gTempDir);
            exit(1);
        }
    }

    if(chmod(gTempDir, S_IRUSR|S_IWUSR|S_IXUSR) < 0) {
        fprintf(stderr, "can't chmod %s", gTempDir);
        exit(1);
    }

    /// 乱数初期化 ///
    srandom(1000);

    /// シグナルを初期化 ///
    set_signal_clear_all();
    signal(SIGWINCH, SIG_IGN);

    /// 環境変数設定 ///
    setenv("SYSCONFDIR", SYSCONFDIR, 1);

    setenv("key_up", string_c_str(string_new_from_int(KEY_UP)), 1);
    setenv("key_right", string_c_str(string_new_from_int(KEY_RIGHT)), 1);
    setenv("key_down", string_c_str(string_new_from_int(KEY_DOWN)), 1);
    setenv("key_left", string_c_str(string_new_from_int(KEY_LEFT)), 1);
    setenv("key_insert", string_c_str(string_new_from_int(KEY_IC)), 1);
    setenv("key_delete", string_c_str(string_new_from_int(KEY_DC)), 1);
    setenv("key_home", string_c_str(string_new_from_int(KEY_HOME)), 1);
    setenv("key_end", string_c_str(string_new_from_int(KEY_END)), 1);
    setenv("key_pageup", string_c_str(string_new_from_int(KEY_PPAGE)), 1);
    setenv("key_pagedown", string_c_str(string_new_from_int(KEY_NPAGE)), 1);
    setenv("key_meta_left", string_c_str(string_new_from_int(KEY_F(13))), 1);
    setenv("key_meta_right", string_c_str(string_new_from_int(KEY_F(14))), 1);
    setenv("key_meta_up", string_c_str(string_new_from_int(KEY_F(15))), 1);
    setenv("key_meta_down", string_c_str(string_new_from_int(KEY_F(16))), 1);
    setenv("key_enter", string_c_str(string_new_from_int(10)), 1);
    setenv("key_backspace", string_c_str(string_new_from_int(KEY_BACKSPACE)), 1);
    setenv("key_f1", string_c_str(string_new_from_int(KEY_F(1))), 1);
    setenv("key_f2", string_c_str(string_new_from_int(KEY_F(2))), 1);
    setenv("key_f3", string_c_str(string_new_from_int(KEY_F(3))), 1);
    setenv("key_f4", string_c_str(string_new_from_int(KEY_F(4))), 1);
    setenv("key_f5", string_c_str(string_new_from_int(KEY_F(5))), 1);
    setenv("key_f6", string_c_str(string_new_from_int(KEY_F(6))), 1);
    setenv("key_f7", string_c_str(string_new_from_int(KEY_F(7))), 1);
    setenv("key_f8", string_c_str(string_new_from_int(KEY_F(8))), 1);
    setenv("key_f9", string_c_str(string_new_from_int(KEY_F(9))), 1);
    setenv("key_f10", string_c_str(string_new_from_int(KEY_F(10))), 1);
    setenv("key_f11", string_c_str(string_new_from_int(KEY_F(11))), 1);
    setenv("key_f12", string_c_str(string_new_from_int(KEY_F(12))), 1);
    
    setenv("key_a", string_c_str(string_new_from_int('a')), 1);
    setenv("key_b", string_c_str(string_new_from_int('b')), 1);
    setenv("key_c", string_c_str(string_new_from_int('c')), 1);
    setenv("key_d", string_c_str(string_new_from_int('d')), 1);
    setenv("key_e", string_c_str(string_new_from_int('e')), 1);
    setenv("key_f", string_c_str(string_new_from_int('f')), 1);
    setenv("key_g", string_c_str(string_new_from_int('g')), 1);
    setenv("key_h", string_c_str(string_new_from_int('h')), 1);
    setenv("key_i", string_c_str(string_new_from_int('i')), 1);
    setenv("key_j", string_c_str(string_new_from_int('j')), 1);
    setenv("key_k", string_c_str(string_new_from_int('k')), 1);
    setenv("key_l", string_c_str(string_new_from_int('l')), 1);
    setenv("key_m", string_c_str(string_new_from_int('m')), 1);
    setenv("key_n", string_c_str(string_new_from_int('n')), 1);
    setenv("key_o", string_c_str(string_new_from_int('o')), 1);
    setenv("key_p", string_c_str(string_new_from_int('p')), 1);
    setenv("key_q", string_c_str(string_new_from_int('q')), 1);
    setenv("key_r", string_c_str(string_new_from_int('r')), 1);
    setenv("key_s", string_c_str(string_new_from_int('s')), 1);
    setenv("key_t", string_c_str(string_new_from_int('t')), 1);
    setenv("key_u", string_c_str(string_new_from_int('u')), 1);
    setenv("key_v", string_c_str(string_new_from_int('v')), 1);
    setenv("key_w", string_c_str(string_new_from_int('w')), 1);
    setenv("key_x", string_c_str(string_new_from_int('x')), 1);
    setenv("key_y", string_c_str(string_new_from_int('y')), 1);
    setenv("key_z", string_c_str(string_new_from_int('z')), 1);

    setenv("key_A", string_c_str(string_new_from_int('A')), 1);
    setenv("key_B", string_c_str(string_new_from_int('B')), 1);
    setenv("key_C", string_c_str(string_new_from_int('C')), 1);
    setenv("key_D", string_c_str(string_new_from_int('D')), 1);
    setenv("key_E", string_c_str(string_new_from_int('E')), 1);
    setenv("key_F", string_c_str(string_new_from_int('F')), 1);
    setenv("key_G", string_c_str(string_new_from_int('G')), 1);
    setenv("key_H", string_c_str(string_new_from_int('H')), 1);
    setenv("key_I", string_c_str(string_new_from_int('I')), 1);
    setenv("key_J", string_c_str(string_new_from_int('J')), 1);
    setenv("key_K", string_c_str(string_new_from_int('K')), 1);
    setenv("key_L", string_c_str(string_new_from_int('L')), 1);
    setenv("key_M", string_c_str(string_new_from_int('M')), 1);
    setenv("key_N", string_c_str(string_new_from_int('N')), 1);
    setenv("key_O", string_c_str(string_new_from_int('O')), 1);
    setenv("key_P", string_c_str(string_new_from_int('P')), 1);
    setenv("key_Q", string_c_str(string_new_from_int('Q')), 1);
    setenv("key_R", string_c_str(string_new_from_int('R')), 1);
    setenv("key_S", string_c_str(string_new_from_int('S')), 1);
    setenv("key_T", string_c_str(string_new_from_int('T')), 1);
    setenv("key_U", string_c_str(string_new_from_int('U')), 1);
    setenv("key_V", string_c_str(string_new_from_int('V')), 1);
    setenv("key_W", string_c_str(string_new_from_int('W')), 1);
    setenv("key_X", string_c_str(string_new_from_int('X')), 1);
    setenv("key_Y", string_c_str(string_new_from_int('Y')), 1);
    setenv("key_Z", string_c_str(string_new_from_int('Z')), 1);

    setenv("key_space", string_c_str(string_new_from_int(32)), 1);
    
    setenv("key_ctrl_space", string_c_str(string_new_from_int(0)), 1);
    setenv("key_ctrl_a", string_c_str(string_new_from_int(1)), 1);
    setenv("key_ctrl_b", string_c_str(string_new_from_int(2)), 1);
    setenv("key_ctrl_c", string_c_str(string_new_from_int(3)), 1);
    setenv("key_ctrl_d", string_c_str(string_new_from_int(4)), 1);
    setenv("key_ctrl_e", string_c_str(string_new_from_int(5)), 1);
    setenv("key_ctrl_f", string_c_str(string_new_from_int(6)), 1);
    setenv("key_ctrl_g", string_c_str(string_new_from_int(7)), 1);
    setenv("key_ctrl_h", string_c_str(string_new_from_int(8)), 1);
    setenv("key_ctrl_i", string_c_str(string_new_from_int(9)), 1);
    setenv("key_ctrl_j", string_c_str(string_new_from_int(10)), 1);
    setenv("key_ctrl_k", string_c_str(string_new_from_int(11)), 1);
    setenv("key_ctrl_l", string_c_str(string_new_from_int(12)), 1);
    setenv("key_ctrl_m", string_c_str(string_new_from_int(13)), 1);
    setenv("key_ctrl_n", string_c_str(string_new_from_int(14)), 1);
    setenv("key_ctrl_o", string_c_str(string_new_from_int(15)), 1);
    setenv("key_ctrl_p", string_c_str(string_new_from_int(16)), 1);
    setenv("key_ctrl_q", string_c_str(string_new_from_int(17)), 1);
    setenv("key_ctrl_r", string_c_str(string_new_from_int(18)), 1);
    setenv("key_ctrl_s", string_c_str(string_new_from_int(19)), 1);
    setenv("key_ctrl_t", string_c_str(string_new_from_int(20)), 1);
    setenv("key_ctrl_u", string_c_str(string_new_from_int(21)), 1);
    setenv("key_ctrl_v", string_c_str(string_new_from_int(22)), 1);
    setenv("key_ctrl_w", string_c_str(string_new_from_int(23)), 1);
    setenv("key_ctrl_x", string_c_str(string_new_from_int(24)), 1);
    setenv("key_ctrl_y", string_c_str(string_new_from_int(25)), 1);
    setenv("key_ctrl_z", string_c_str(string_new_from_int(26)), 1);
    setenv("key_escape", string_c_str(string_new_from_int(27)), 1);
    setenv("key_tab", string_c_str(string_new_from_int(9)), 1);

    setenv("key_0", string_c_str(string_new_from_int('0')), 1);
    setenv("key_1", string_c_str(string_new_from_int('1')), 1);
    setenv("key_2", string_c_str(string_new_from_int('2')), 1);
    setenv("key_3", string_c_str(string_new_from_int('3')), 1);
    setenv("key_4", string_c_str(string_new_from_int('4')), 1);
    setenv("key_5", string_c_str(string_new_from_int('5')), 1);
    setenv("key_6", string_c_str(string_new_from_int('6')), 1);
    setenv("key_7", string_c_str(string_new_from_int('7')), 1);
    setenv("key_8", string_c_str(string_new_from_int('8')), 1);
    setenv("key_9", string_c_str(string_new_from_int('9')), 1);
        
    setenv("key_exclam", string_c_str(string_new_from_int('!')), 1);
    setenv("key_dquote", string_c_str(string_new_from_int('"')), 1);
    setenv("key_sharp", string_c_str(string_new_from_int('#')), 1);
    setenv("key_dollar", string_c_str(string_new_from_int('$')), 1);
    setenv("key_percent", string_c_str(string_new_from_int('%')), 1);
    setenv("key_and", string_c_str(string_new_from_int('&')), 1);
    setenv("key_squote", string_c_str(string_new_from_int('\'')), 1);
    setenv("key_lparen", string_c_str(string_new_from_int('(')), 1);
    setenv("key_rparen", string_c_str(string_new_from_int(')')), 1);
    setenv("key_tilda", string_c_str(string_new_from_int('~')), 1);
    setenv("key_equal", string_c_str(string_new_from_int('=')), 1);
    setenv("key_minus", string_c_str(string_new_from_int('-')), 1);
    setenv("key_cup", string_c_str(string_new_from_int('^')), 1);
    setenv("key_vbar", string_c_str(string_new_from_int('|')), 1);
    setenv("key_backslash", string_c_str(string_new_from_int('\\')), 1);
    setenv("key_atmark", string_c_str(string_new_from_int('@')), 1);
    setenv("key_bapostrophe", string_c_str(string_new_from_int('`')), 1);
    setenv("key_lcurly", string_c_str(string_new_from_int('{')), 1);
    setenv("key_lbrack", string_c_str(string_new_from_int('[')), 1);
    setenv("key_plus", string_c_str(string_new_from_int('+')), 1);
    setenv("key_semicolon", string_c_str(string_new_from_int(';')), 1);
    setenv("key_star", string_c_str(string_new_from_int('*')), 1);
    setenv("key_colon", string_c_str(string_new_from_int(':')), 1);
    setenv("key_rcurly", string_c_str(string_new_from_int('}')), 1);
    setenv("key_rbrack", string_c_str(string_new_from_int(']')), 1);
    setenv("key_lss", string_c_str(string_new_from_int('<')), 1);
    setenv("key_comma", string_c_str(string_new_from_int(',')), 1);
    setenv("key_gtr", string_c_str(string_new_from_int('>')), 1);
    setenv("key_dot", string_c_str(string_new_from_int('.')), 1);
    setenv("key_slash", string_c_str(string_new_from_int('/')), 1);
    setenv("key_qmark", string_c_str(string_new_from_int('?')), 1);
    setenv("key_underbar", string_c_str(string_new_from_int('_')), 1);
    
    setenv("nometa", string_c_str(string_new_from_int(0)), 1);
    setenv("meta", string_c_str(string_new_from_int(1)), 1);

    setenv("ma_reverse", string_c_str(string_new_from_int(kCAReverse)), 1);
    setenv("ma_bold", string_c_str(string_new_from_int(kCABold)), 1);
    setenv("ma_underline", string_c_str(string_new_from_int(kCAUnderline)), 1);
    setenv("ma_white", string_c_str(string_new_from_int(kCAWhite)), 1);
    setenv("ma_blue", string_c_str(string_new_from_int(kCABlue)), 1);
    setenv("ma_cyan", string_c_str(string_new_from_int(kCACyan)), 1);
    setenv("ma_green", string_c_str(string_new_from_int(kCAGreen)), 1);
    setenv("ma_yellow", string_c_str(string_new_from_int(kCAYellow)), 1);
    setenv("ma_magenta", string_c_str(string_new_from_int(kCAMagenta)), 1);
    setenv("ma_red", string_c_str(string_new_from_int(kCARed)), 1);


    /// オプション処理 ///
    char send_cmd[BUFSIZ];
    strcpy(send_cmd, "");

    char send_cmd2[BUFSIZ];
    strcpy(send_cmd2, "");
    int opt;
    while((opt = getopt(argc, argv, "e:c:")) != -1) {
        switch(opt) {
        case 'e':
            strcpy(send_cmd, optarg);
            break;

        case 'c':
            strcpy(send_cmd2, optarg);
            break;

        case ':':
            printf("option needs a value\n");
            exit(1);
            break;

        case '?':
            printf("usage mfiler3 [-c command]\n\n");
            printf("-c : send a command to running mfiler3\n");
            exit(1);
            break;
        }
    }

    /// 環境変数初期化 ///
    setenv("#", "-1", 1);

    /// オプションで指定されているならシェルで実行する ///
    if(strcmp(send_cmd2, "") != 0) {
        if(gKanjiCode == kUtf8) setlocale(LC_CTYPE, "ja_JP.UTF-8");
        set_signal_subshell();
        //tcsetpgrp(0, getpid());
        //debug_init();
        shell_init_subshell();

        int rcode;
        if(!isatty(0) || !isatty(1)) {
            int pipefds[2];

            if(pipe(pipefds) < 0) {
                perror("pipe");
                exit(1);
            }

            rcode = shell(send_cmd2, NULL, pipefds[1], FALSE, TRUE);
            
            string_obj* str = string_new("");
            char buf[BUFSIZ];
            while(1) {
                int size = read(pipefds[0], buf, BUFSIZ-1);
                if(size == 0) {
                    break;
                }
                if(size < 0) {
                    fprintf(stderr, "mfiler -c (%s) failed (read)", send_cmd2);
                    close(pipefds[0]);
                    return 1;
                }

                buf[size] = 0;

                string_push_back(str, buf);
            }
            close(pipefds[0]);

            printf("%s", string_c_str(str));
        }
        else {
            rcode = shell(send_cmd2, NULL, -1, FALSE, TRUE);
        }

        //debug_final();

        return rcode;
    }

    /// オプションで指定されているなら動いているmfiler3にメッセージを送る  ///
    if(strcmp(send_cmd, "") != 0) {
        DIR* dir = opendir(gTempDir);
        if(dir == NULL) {
            fprintf(stderr, "Tempolary directory doesn't exist");
            exit(1);
        }

        struct dirent* entry;
        while(entry = readdir(dir)) {
            if(strstr(entry->d_name, "socket")) {
                char soc_name[PATH_MAX];
                sprintf(soc_name, "%s/%s", gTempDir, entry->d_name);

                int soc = socket(AF_UNIX, SOCK_DGRAM, 0);
                if(soc < 0) {
                    perror("socket");
                    return 1;
                }

                struct sockaddr_un addr;
                addr.sun_family = AF_UNIX;
                strcpy(addr.sun_path, soc_name);
                if(connect(soc, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
                    write(soc, send_cmd, strlen(send_cmd)+1);
                }

                close(soc);
            }
        }
        
        return 0;
    }

    /// 標準入力と出力が制御端末かどうか確認 ///
    if(!isatty(0) || !isatty(1)) {
        fprintf(stderr, "standard input is not a tty\n");
        return 1;
    }

    /// フォワグランドジョブになるまでループ ///
    int status;
    pid_t pgrp;
    while((status = tcgetpgrp(0)) >= 0) {
        if(status == (pgrp = getpgrp())) {
            break;
        }
        kill(- pgrp, SIGTTIN);
    }

    /// 新しいセッションを始めてプロセスグループリーダーになる ///
    setsid();
    pid_t shell_pgrp = getpid();
    setpgid(shell_pgrp, shell_pgrp);

    /// 端末の制御を握る ///
    tcsetpgrp(0, shell_pgrp);

    /// モジュールの初期化 ///
    cmdline_init();
    cmdline_completion_init();
    menu_init();
    //debug_init();
    filer_init();
    mcurses_init();
    kanji_init();
    shell_init();

    /// ソケットの初期化 ///
    char soc_name[PATH_MAX];
    int soc;
    sprintf(soc_name, "%s/socket%d", gTempDir, getpid());
    
    if(access(soc_name, F_OK) == 0) {
        unlink(soc_name);
    }
    
    soc = socket(AF_UNIX, SOCK_DGRAM, 0);
    if(soc < 0) {
        perror("socket");
        return 1;
    }
    
    struct sockaddr_un addr;
    addr.sun_family = AF_UNIX;
    strcpy(addr.sun_path, soc_name);
    if(bind(soc, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
        perror("bind");
        return 1;
    }

    /// シグナルを設定 ///
    set_signal_mfiler3();

    /// ランタイムスクリプト実行 ///
    read_rc_file();
    cmdline_rehash();

    /// モジュールの初期化(ランタイムスクリプト以降の方が良い奴 ///
    isearch_init();

    /// localeの設定 ///
    if(gKanjiCode == kUtf8) setlocale(LC_CTYPE, "ja_JP.UTF-8");

    /// ディレクトリを作成 ///
    filer_new_dir(mygetcwd());
    filer_new_dir(mygetcwd());

    filer_activate(0);
    
    /// カーシスの起動 ///
    minitscr();
    mclear_immediately();

    /// atexit登録 ///
    atexit(atexit_fun);
    
    /// メインループ ///
    gMainLoop = TRUE;

    fd_set mask;
    fd_set read_ok;

    FD_ZERO(&mask);
    FD_SET(0, &mask);
    FD_SET(soc, &mask);

    while(gMainLoop) {
        /// 描写 ///
        mclear();
        view();
        mrefresh();

        /// selectで入力待ち ///
        read_ok = mask;
        if(!mkbuf_exist()) {     // キーボードのバッファがないならselect
            if(vector_size(gJobs) > 0) {
                struct timeval tv;
                tv.tv_sec = 5;
                tv.tv_usec = 0;

                select(soc + 1, &read_ok, NULL, NULL, &tv);
            }
            else {
                select(soc + 1, &read_ok, NULL, NULL, NULL);
            }
        }

        /// キー処理 ///
        if(FD_ISSET(0, &read_ok) || mkbuf_exist()) {    // キー入力があったか、キーボードのバッファがあるなら
            /// コマンドライン入力中 ///
            if(gCmdLineActive) {
                while(1) {
                    /// 漢字入力のためselectで入力を見張る ///
                    struct timeval tv;
                    tv.tv_sec = 0;
                    tv.tv_usec = 0;

                    read_ok = mask;

                    select(1, &read_ok, NULL, NULL, &tv);

                    /// キー入力中 ///
                    if(FD_ISSET(0, &read_ok) || mkbuf_exist()) {
                        int meta;
                        int key = mgetch(&meta);

                        if(key == -1) {
                        }
                        else if(meta) {
                            input(1, key);
                        }
                        /// normal key ///
                        else {
                           input(0, key);
                        }
                    }
                    /// キー入力が終わった ///
                    else {
                        break;
                    }
                }
            }

            /// その他 ///
            else {
                int meta;
                int key = mgetch_nonblock(&meta);

                /// ESCキー ///
                if(key == -1) {
                }
                else if(meta) {
                    input(1, key);
                }
                
                /// meta key ///
                else if(key >= kKeyMetaFirst && key <= kKeyMetaFirst+127) {
                    input(1, key - kKeyMetaFirst);
                }

                /// normal key ///
                else {
                   input(0, key);
                }
            }
        }

        /// UNIX Domainソケット処理 ///
        else if(FD_ISSET(soc, &read_ok)) {
            char buf[BUFSIZ];
            
            struct sockaddr_un caddr;
            socklen_t caddr_len;
            int len = recvfrom(soc, buf, BUFSIZ, 0, (struct sockaddr *)&caddr, &caddr_len);
            shell(buf, NULL, -1, FALSE, FALSE);
        }

        /// ジョブ処理 ///
        shell_wait_backgroud_job();
    }

    /// カーシス終わり ///
    mmove_immediately(0,0);
    mclear_immediately();
    mendwin();

    /// モジュール解放 ///
    isearch_final();
    mcurses_final();
    kanji_final();
    //debug_final();

    /// フック呼び出し ///
    shell("%Q%h atexit_fun", "atexit_fun", -1, FALSE, TRUE);

    /// シグナルを初期化 ///
    set_signal_clear_all();
    
    /// 全てのジョブをkill ///
    shell_kill_all_jobs();

    /// ソケット解放 ///
    unlink(soc_name);

    return 0;
}
